# Copyright (c) 2025 Mobius Logic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class to create alphanumeric strings.

Generates random strings which each character being a random letter or digit.
Letters include both upper or lower case. The desired string length must be
specified.

Typical usage example:
  seed = 123
  rng = random.Random(seed)
  obj = AlphaNumNickNamer(length=10)
  random_str = obj.get_value_from_generator(rng)

Or, to use with NickNamer class:
  seed = 123
  obj = NickNamer(seed, AlphaNumNickNamer, length=10)
  random_str = obj.get_value()
"""

import random
import string

from typing_extensions import override

from nicknamer.nicknamer_base import RandomWithChoice, _BaseNickNamer


class AlphaNumNickNamer(_BaseNickNamer):
    """Generate random alphanumeric sequence.

    Sequence if a defined length string with characters taken from a-zA-Z0-9.
    """

    def __init__(
        self,
        length: int,
    ) -> None:
        """Initialize the object.

        Args:
          length: the desired length of the string

        Raises:
          ValueError: If length is less than 1.
        """
        if length < 1:
            err_msg = "'length' must be >= 1"
            raise ValueError(err_msg)
        self._length = length

    @override
    def get_value_from_generator(self, rng: RandomWithChoice | random.Random) -> str:
        characters = list(string.ascii_letters + string.digits)

        try:
            # Try the numpy version
            vals = rng.choice(
                characters,
                size=self._length,  # pyright: ignore[reportCallIssue]
            )
            return "".join(vals)
        except TypeError:
            # try the random.Random version
            vals = rng.choices(  # pyright: ignore[reportAttributeAccessIssue]
                characters,
                k=self._length,
            )
            return "".join(vals)



# Copyright (c) 2025 Mobius Logic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base classes for NickNamer.

These provide the base behavior for the nick naming classes.
The actual string creation behavior is controlled by the other classes.
See each class for details.

The NickNamer class provides an optional framework for tracking the state of
the generators.

Pseudocode of typical usage:
  seed = 123
  obj = NickNamer(seed, NICK_NAMER_CLASS, **CLASS_SPECIFIC_CONFIG_VALUES)
  random_str = obj.get_value()
"""

import random
from abc import ABC, abstractmethod
from typing import Any, Protocol


class RandomWithChoice(Protocol):
    """Matches any object that has a suitable choice method."""

    def choice(self, seq: Any, size: int = 1) -> Any:
        """Select a random item from the sequence."""
        ...


class _BaseNickNamer(ABC):
    """Base class for nick namer objects."""

    @abstractmethod
    def get_value_from_generator(self, rng: RandomWithChoice | random.Random) -> str:
        """Return a string created using the given random generator.

        This uses the passed random number generator.

        The value, format, style, etc... of the returned string are controlled
        by the specific class being used. See the class definition for details.

        Args:
          rng: random generator to use in creation

        Returns:
          the randomly generated id
        """


class NickNamer:
    """Generate Random IDs."""

    def __init__(
        self,
        seed: int | float | str | bytes | bytearray | None,
        generator_class: type[_BaseNickNamer],
        **kwargs: Any,
    ) -> None:
        """Initialize the NickNamer object.

        This will internally create an object to generate ids using the given
        `generator_class`. That object will be passed `kwargs` for construction,
        so see the relevant class for details.

        Args:
          seed: the seed for the random number generator
          generator_class: class to use for making the values
          kwargs: values passed to `generator_class` as parameters.
            See the specific class for required and optional values.
        """
        self._rng = random.Random(seed)
        self._name_generator = generator_class(**kwargs)

    def get_value(self) -> str:
        """Return a random id using the internal random generator.

        Returns:
          The random id generated from the object.
        """
        return self._name_generator.get_value_from_generator(self._rng)

    def get_value_from_generator(self, rng: RandomWithChoice | random.Random) -> str:
        """Return a random id using the given random generator.

        Args:
          rng: a random generator that has a choice() method to call

        Returns:
          The random id generated from the object.
        """
        return self._name_generator.get_value_from_generator(rng)



# Copyright (c) 2025 Mobius Logic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Run tests on the word generator."""

import random

import pytest

import nicknamer

# format string, expected parsed list
format_examples = [
    ("<noun>", ["<noun>"]),
    ("<noun><noun>", ["<noun>", "<noun>"]),
    ("<noun>.<noun>", ["<noun>", ".", "<noun>"]),
    (
        "<noun>, <verb>, and <noun>!",
        ["<noun>", ", ", "<verb>", ", and ", "<noun>", "!"],
    ),
    ("<adjective>_<noun>_<verb>", ["<adjective>", "_", "<noun>", "_", "<verb>"]),
    ("", []),
]


@pytest.mark.parametrize(("format_string", "output_list"), format_examples)
def test_formatter(format_string: str, output_list: list[str]) -> None:
    """Test that format parsing works as expected."""
    obj = nicknamer.WordNickNamer(format_str=format_string)
    actual_parsing = obj._parsed_format  # noqa: SLF001
    if actual_parsing != output_list:
        err_msg = (
            "WordNickNamer: parsed format does not match expected: ",
            f"{actual_parsing} != {output_list}",
        )
        raise AssertionError(err_msg)


result_examples_seed123 = [
    ("<noun>", "Articulateness"),
    ("<noun><noun>", "ArticulatenessEbrillade"),
    ("<noun>.<noun>", "Articulateness.Ebrillade"),
    ("<noun>, <verb>, and <noun>!", "Articulateness, Embosser, and Bishopdom!"),
    ("<adjective>_<noun>_<verb>", "Araneiform_Ebrillade_Bonesetter"),
    ("", ""),
]


@pytest.mark.parametrize(("format_string", "output_value"), result_examples_seed123)
def test_word_generator(format_string: str, output_value: str) -> None:
    """Test that words are generated as expected."""
    seed = 123
    obj = nicknamer.NickNamer(
        seed=seed,
        generator_class=nicknamer.WordNickNamer,
        format_str=format_string,
    )
    actual_value = obj.get_value()
    if actual_value != output_value:
        err_msg = (
            "WordNickNamer: output does not match expected: ",
            f"{actual_value} != {output_value}",
        )
        raise AssertionError(err_msg)


@pytest.mark.parametrize(("format_string", "output_value"), result_examples_seed123)
def test_word_generator_external(format_string: str, output_value: str) -> None:
    """Test that words are generated as expected using external generator."""
    wrong_seed = 456  # this will not give the expected value
    obj = nicknamer.NickNamer(
        seed=wrong_seed,
        generator_class=nicknamer.WordNickNamer,
        format_str=format_string,
    )
    wrong_value = obj.get_value()  # should be wrong because the seed is wrong

    # ignore empty output_value, which will match regardless of seed
    if wrong_value == output_value and output_value:
        err_msg = (
            "WordNickNamer: output is wrong: ",
            f"{wrong_value} == {output_value}  ;{format_string};",
        )
        raise AssertionError(err_msg)

    right_seed = 123
    rng = random.Random(right_seed)
    right_value = obj.get_value_from_generator(rng)  # uses correct generator
    if right_value != output_value:
        err_msg = (
            "WordNickNamer: output incorrect with external generator ",
            f"{right_value} != {output_value}",
        )
        raise AssertionError(err_msg)



# Copyright (c) 2025 Mobius Logic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Run tests on the word generator."""

import random

import pytest

import nicknamer


def test_alphanumeric_length() -> None:
    """Test that lengths are handled correctly."""
    right_seed = 123
    rng = random.Random(right_seed)

    for length in range(1, 5):
        obj = nicknamer.AlphaNumNickNamer(length=length)
        if len(obj.get_value_from_generator(rng)) != length:
            err_msg = "Wrong length returned from AlphaNumNickNamer"
            raise AssertionError(err_msg)

    for length in [-10, 0]:
        with pytest.raises(ValueError, match="'length' must be >= 1"):
            _ = nicknamer.AlphaNumNickNamer(length=length)


# length, seed, example result1, examplt result2
result_examples = [
    (12, 123, "dfzg3cHu0juu", "paBfLetB4fiW"),
    (30, 324, "KS7DHZeqlvwNy1JX0CimS0Oda6MVZV", "XaRucukpUt9Hjq09VKNhsVoffdi0rN"),
    (6, 1, "i0VpEB", "OWfbZA"),
]


@pytest.mark.parametrize(
    ("length", "seed", "example_result1", "example_result2"),
    result_examples,
)
def test_alphanumeric_generator(
    length: int,
    seed: int,
    example_result1: str,
    example_result2: str,
) -> None:
    """Test that words are generated as expected."""
    obj = nicknamer.NickNamer(
        seed=seed,
        generator_class=nicknamer.AlphaNumNickNamer,
        length=length,
    )
    for expected in [example_result1, example_result2]:
        actual_value = obj.get_value()
        if actual_value != expected:
            err_msg = (
                "AlphaNumNickNamer: output does not match expected: ",
                f"{actual_value} != {expected}",
            )
            raise AssertionError(err_msg)



# Copyright (c) 2025 Mobius Logic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class to create random words.

Generates random strings with defined format that contains specific parts of
speech. Example format: "<noun>.<verb>" gives a noun and verb, separated by
a period. For example 'truck.sing'.

Typical usage example:
  seed = 123
  rng = random.Random(seed)
  format_str = "<noun>.<verb>"
  obj = WordNickNamer(format_str)
  random_str = obj.get_value_from_generator(rng)

Or, to use with NickNamer class:
  seed = 123
  format_str = "<noun>.<verb>"
  obj = NickNamer(seed, WordNickNamer, format_str=format_str)
  random_str = obj.get_value()
"""


import importlib.resources
import random
import re
from typing import ClassVar

from typing_extensions import override

from nicknamer.nicknamer_base import RandomWithChoice, _BaseNickNamer


class WordNickNamer(_BaseNickNamer):
    """Generate random word sequence."""

    _wordlist: ClassVar[dict[str, list[str]]] = {}

    def __init__(
        self,
        format_str: str,
    ) -> None:
        """Initialize the object.

        `format_str` is a string describing which random words will be used and what
        the format of the final string will be. For example, "<noun>.<noun>" will
        give two random nounds separated by a '.'. For example:
        bag.line, store.peanut, car.house

        valid parts of speech are: adjective, noun, and verb.

        Parameters
        ----------
        format_str
            Format of the random str
        """
        # create wordlist on first instance
        if not WordNickNamer._wordlist:
            WordNickNamer._fill_wordlist()

        self._format = format_str
        self._parsed_format = self._get_parsed_format()

    def _get_parsed_format(self) -> list[str]:
        """Parse the format string into a sequence parts.

        This is done to allow the format to be used without using a regular
        expression every time.

        This will take "<noun>, <verb>, and <noun>!" and return:
        ["<noun>", ", ", "<verb>", ", and ", "<noun>", "!"]
        When generating a new random value, <noun>, <verb>, and <noun> will be
        replaced with random words of the appropriate type.

        Returns:
          A list of the format sequence parts
        """
        # format for a word, example <noun>
        pattern = re.compile(r"(<.*?>)")
        parsed_format = []
        last_pos = 0

        for match in pattern.finditer(self._format):
            # everything before the match, the matched word type
            parsed_format.extend(
                (self._format[last_pos : match.start()], match.group(1)),
            )
            last_pos = match.end()

        # The rest of the text after the last match
        parsed_format.append(self._format[last_pos:])

        # remove '' from the list
        return [i for i in parsed_format if i]

    @classmethod
    def _fill_wordlist(cls) -> None:
        """Read the word lists from the resource files."""
        for word_type in ["adjective", "noun", "verb"]:
            filename = f"{word_type}.txt"

            with importlib.resources.open_text("nicknamer.resources", filename) as f:
                cls._wordlist[f"<{word_type}>"] = [
                    word.strip() for word in f.read().strip().split(",")
                ]

    @override
    def get_value_from_generator(self, rng: RandomWithChoice | random.Random) -> str:
        return "".join(
            rng.choice(self._wordlist[part]) if part in self._wordlist else part
            for part in self._parsed_format
        )
